import sqlite3
from collections import namedtuple
from datetime import datetime
from typing import Generator

from NotAOrm.condition import Condition
from NotAOrm.sql import option, order


class Query:
    def __init__(self, table_name, path_database):
        self.table_name = table_name
        self._conn = sqlite3.connect(path_database, detect_types=sqlite3.PARSE_DECLTYPES)
        sqlite3.register_converter("BOOLEAN", lambda v: v.decode() == 'True')

    def _get_table_object(self, descriptions: tuple):
        return namedtuple(self.table_name, [desc[0] for desc in descriptions])

    @staticmethod
    def _append_option(query: str, **kwargs):
        for key, value in kwargs.items():
            if not hasattr(option, key.upper()):
                raise NotImplementedError('Option not implement')

            query += getattr(option, key.upper()).format(value)
        return query

    def _fetch(self, query: str, *args, **kwargs):
        columns = kwargs.pop('columns')
        if columns is not '*':
            columns = ','.join(repr(c) for c in columns) if type(columns) is list else repr(columns)

        full_query = self._append_option(query, **kwargs).replace('COLUMNS_NAME', columns)
        res = self.exec(full_query, *args, commit=False)
        table_obj = self._get_table_object(res.description)

        return res, table_obj

    def _fetch_all(self, query: str, *args, **kwargs):
        res, table_obj = self._fetch(query, *args, **kwargs)

        for items in res.fetchall():
            yield table_obj(*items)

    def _fetch_one(self, query: str, *args, **kwargs):
        res, table_obj = self._fetch(query, *args, **kwargs)

        fetch = res.fetchone()
        if fetch is not None:
            return table_obj(*fetch)

    def exec(self, query: str, *args, commit=True):
        query = query.replace('TABLE_NAME', self.table_name)
        res = self._conn.execute(query, args)

        if commit:
            self._conn.commit()

        return res


class Change(Query):
    def update(self, condition, **columns) -> sqlite3.Cursor:
        columns_to_set = ",".join(f'{key} = ?' for key in columns.keys())
        values = list(columns.values()) + condition.values

        return self.exec(order.UPDATE.format(columns_to_set, condition.left_side), *values)

    def insert(self, **columns) -> sqlite3.Cursor:
        keys = ",".join(columns.keys())
        values = ','.join('?' * len(columns.values()))

        return self.exec(order.INSERT.format(keys, values), *columns.values())

    def delete(self, condition: Condition, commit=False) -> sqlite3.Cursor:
        return self.exec(order.DELETE.format(condition.left_side), *condition.values, commit=commit)


class Show(Query):
    def all(self, columns='*', **options) -> Generator:
        return self._fetch_all(order.SELECT_ALL, columns=columns, **options)

    def filter(self, condition: Condition, columns='*', **options) -> Generator:
        return self._fetch_all(
            order.SELECT_WHERE.format(condition.left_side),
            *condition.values,
            columns=columns,
            **options
        )

    def get(self, condition: Condition, columns='*', **options) -> tuple:
        return self._fetch_one(
            order.SELECT_WHERE.format(condition.left_side),
            *condition.values,
            columns=columns,
            **options
        )

    def first(self, columns='*') -> tuple:
        return self._fetch_one(order.SELECT_ALL, columns=columns, order_by_asc=f'{self.table_name}.OID', limit=1)

    def last(self, columns='*') -> tuple:
        return self._fetch_one(order.SELECT_ALL, columns=columns, order_by_desc=f'{self.table_name}.OID', limit=1)



from NotAOrm.condition import Condition
from NotAOrm.operator_enum import Comparator
from NotAOrm.sql import creation
from NotAOrm.table import Row, MathFunction


class _Generic(Row):
    prefix = ''

    def get_sql_creation(self):
        return super().get_sql_creation().format(prefix=self.prefix)


class Int(_Generic):
    prefix = creation.INTEGER

    def __init__(self, row_name: str, primary_key=False, **kwargs):
        super().__init__(row_name, **kwargs)
        self.primary_key = primary_key

    def get_sql_creation(self):
        if self.primary_key:
            self.prefix += f' {creation.PRIMARY_KEY} {creation.AUTO_INCREMENT}'
        return super().get_sql_creation().format(prefix=self.prefix)

    @property
    def sum(self):
        return MathFunction('SUM', self.__repr__(), f'sum_{self.row_name}')


class Float(_Generic):
    prefix = creation.FLOAT

    @property
    def sum(self):
        return MathFunction('SUM', self.__repr__(), f'sum_{self.row_name}')


class Varchar(_Generic):
    def __init__(self, row_name, length=255, **kwargs):
        super().__init__(row_name, **kwargs)
        self.prefix = creation.VARCHAR.format(length)

    def is_like(self, other):
        return Condition(self.__repr__(), Comparator.str, '?', [other])


class Text(_Generic):
    def __init__(self, row_name, length=5000, **kwargs):
        super().__init__(row_name, **kwargs)
        self.prefix = creation.TEXT.format(length)


class Date(_Generic):
    prefix = creation.DATE
    now = 'CURRENT_DATE '

    def __init__(self, row_name: str, default=None, **kwargs):
        kwargs['default'] = self.now if default == 'now' else f'`{default}`'
        super().__init__(row_name, **kwargs)


class Datetime(Date):
    prefix = creation.TIMESTAMP
    now = 'CURRENT_TIMESTAMP '


class Bool(_Generic):
    prefix = creation.BOOLEAN

    def __init__(self, row_name: str, default=None, **kwargs):
        if default is not None:
            kwargs['default'] = 1 if default else 0
        super().__init__(row_name, **kwargs)



from typing import Union

from NotAOrm.operator_enum import Conditional, Comparator
from NotAOrm.sql.order import WHERE


class Condition:
    def __init__(self, left: str, operator: Union[Conditional, Comparator], right: str, values: list):
        self.left_side = WHERE.format(left, operator.value, right)
        self.values = values

    def __and__(self, other: 'Condition'):
        self.values.extend(other.values)
        return Condition(self.left_side, Conditional.and_, other.left_side, self.values)

    def __or__(self, other: 'Condition'):
        self.values.extend(other.values)
        return Condition(self.left_side, Conditional.or_, other.left_side, self.values)



from typing import Sequence

import NotAOrm
from NotAOrm.condition import Condition
from NotAOrm.operator_enum import Comparator
from NotAOrm.query import Show, Change
from NotAOrm.sql import order, creation


class Table:
    def __init__(self, table_name: str, table_row: Sequence['Row']):
        self.table_name = table_name
        self.show = Show(table_name, NotAOrm.database)
        self.change = Change(table_name, NotAOrm.database)
        self._table_row = table_row
        self.pk = Row('OID')

        for row in table_row:
            row.table_name = table_name
            setattr(self, row.row_name, row)

    def create(self):
        row_sql = ', '.join(row.get_sql_creation() for row in self._table_row)
        full_sql = creation.CREATE_TABLE.format(row_sql)
        self.change.exec(full_sql)

        for row_unique in [r for r in self._table_row if r.unique]:
            self.change.exec(creation.UNIQUE.format(row_name=row_unique.row_name))

    def __str__(self):
        return self.table_name


class Row:
    def __init__(self, row_name: str, not_null=False, unique=False, default=None):
        self.row_name = row_name.replace(' ', '_')
        self.not_null = not_null
        self.unique = unique
        self.table_name = None
        self.default = default

    def __eq__(self, other):
        return Condition(self.__repr__(), Comparator.equ, '?', [other])

    def __ne__(self, other):
        return Condition(self.__repr__(), Comparator.diff, '?', [other])

    def __lt__(self, other):
        return Condition(self.__repr__(), Comparator.inf, '?', [other])

    def __le__(self, other):
        return Condition(self.__repr__(), Comparator.infEq, '?', [other])

    def __gt__(self, other):
        return Condition(self.__repr__(), Comparator.sup, '?', [other])

    def __ge__(self, other):
        return Condition(self.__repr__(), Comparator.supEq, '?', [other])

    def __repr__(self):
        return f'{self.table_name}.{self.row_name}' if self.table_name else self.row_name

    @property
    def count(self):
        return MathFunction('COUNT', self.__repr__(), f'count_{self.row_name}')

    def get_sql_creation(self):
        sql = f'{self.row_name} {{prefix}} '
        if self.default is not None:
            sql += creation.DEFAULT.format(self.default)
        if self.not_null:
            sql += f'{creation.NOT_NULL}'
        return sql


class MathFunction:
    def __init__(self, math, row, label):
        self.math = math
        self.row = row
        self.label = label

    def __repr__(self):
        return getattr(order, self.math).format(self.row, self.label)



from enum import Enum


class Comparator(Enum):
    str = 'LIKE'
    equ = '=='
    sup = '>'
    supEq = '>='
    inf = '<'
    infEq = '<='
    diff = '!='


class Conditional(Enum):
    and_ = ' AND '
    or_ = ' OR '
